{{ sdram-video-20MHz.spin }}
CON
  EMPTY = $FFFF  ' empty queue slot flag
  COMM_COLOURS = 11
  COMM_SPAN    = 22
  COMM_CHAR    = 33

  WAITCYCLES = 16
  WIDTH = 800
  HEIGHT = 480
  GAP = 100      ' pixels per hblank and vblank

VAR
  long qsize
  long qp                  ' insert pointer into queue
  long queue
  long cog
  long args [4]
  long lockid


PUB  Start (qs, q, offsaddr, lid)
  qsize := qs
  queue := q
  lockid := lid
  init_queue

{ Launch SDRAM cog }
  args[0] := queue	   ' pass the queue address,
  args[1] := qsize         ' and size,
  args[2] := offsaddr      ' and the address in SDRAM of first scanline
  args[3] := lockid
  cog := 1 + cognew (@entry, @args) 'Launch new cog
  return cog

PUB  Stop
  if cog <> 0
    cogstop (cog-1)
    cog := 0

{ queue handling code }
PUB init_queue | i
  qp := 0
  repeat i from 0 to QSIZE-1
    word [queue] [i*4+1] := EMPTY

PUB  drain_queue | i
  repeat i from 0 to QSIZE-1
    repeat until word [queue] [i*4+1] == EMPTY


PUB  set_colours (fore, back)
  enqueue (COMM_COLOURS, fore, back, fore)

PUB  draw_span (x, y, wid)
  enqueue (COMM_SPAN, x, y, wid)

PUB  draw_glyph (x, y, chr)
  enqueue (COMM_CHAR, x, y, chr)

PUB  fill_rect (l, t, r, b) | i, j, w
  if l < r and t < b
    w := r-l+1
    repeat i from t to b
      enqueue (COMM_SPAN, l, i, w)

PUB  draw_rect (l, t, r, b) | i, j, w
  if l < r and t < b
    w := r-l+1
    enqueue (COMM_SPAN, l, t, w)
    repeat i from t+1 to b-1
      enqueue (COMM_SPAN, l, i, 1)
      enqueue (COMM_SPAN, r, i, 1)
    enqueue (COMM_SPAN, l, b, w)

PUB  border_rect (l, t, r, b)
  fill_rect (l, t, r, b)
  set_colours (0,0)
  draw_rect (l, t, r, b)


PUB  do_circle (x, y, radius) | f, dx, dy, x1, y1   ' circle drawing code
  f := 1 - radius
  dx := 1
  dy := -2 * radius
  x1 := 0
  y1 := radius
  circle_slice (x-radius, x+radius, y)
  repeat
    if f => 0
      y1 --
      dy += 2
      f += dy
    x1 ++
    dx += 2
    f += dx
    circle_slice (x-x1, x+x1, y+y1)
    circle_slice (x-x1, x+x1, y-y1)
    circle_slice (x-y1, x+y1, y+x1)
    circle_slice (x-y1, x+y1, y-x1)
  until x1 => y1

PRI  circle_slice (xl, xh, y) | w
  w := xh-xl+1
  if w > 0                 ' the renderer isn't bug free, short spans go wrong!
    draw_span (xl, y, w)

  repeat until word [queue] [qp*4+1] == EMPTY   ' wait till next slot empty

PRI  enqueue (comm, x, y, z) | emp
  repeat
    repeat until not lockset (lockid)
    emp := word [queue] [qp*4+1]
    if emp == EMPTY
      word [queue] [qp*4] := x
      word [queue] [qp*4+3] := y
      word [queue] [qp*4+2] := z
      word [queue] [qp*4+1] := comm                   ' 4 word fields: colour, x, width, y
      qp ++
      if qp => qsize
        qp -= qsize
      lockclr (lockid)
      return
    lockclr (lockid)


DAT
		ORG	0

entry           mov     parm, PAR
                rdlong  queue_beg, parm
                add     parm, #4
                rdlong  queue_end, parm         ' actually the length in entries
                shl     queue_end, #3           ' convert to byte length
                add     queue_end, queue_beg
                add     parm, #8
		rdlong  lock, parm
                sub     parm, #4
		rdlong	scanbase, parm		' parm left pointing at scanbase


                mov     OUTA, #0
                or	DIRA, ADDR_MASK
                or      DIRA, CKEMASK           ' ensure CKE driven LOW
		or	DIRA, LCLKMASK          ' and LCLK

                mov	FRQA, CLOCK_FREQ
                mov     PHSA, PHASE
		movs	CTRA, pin_SCLK
		movi	CTRA, #%0_00010_101	' PLL / 4, so 20MHz

                mov	FRQB, CLOCK_FREQ
                mov     PHSB, PHASE
		movs	CTRB, pin_DCLK
		movi	CTRB, #%0_00010_101

		mov	VSCL, V4096
		mov	VCFG, vconfig_go
		WAITVID	disabled, #0		' force the video hardware to a known state
                or      DIRA, VIDMASK
                or      DIRA, SCLK_MASK		' now clocking in inactive state
                or      DIRA, DCLK_MASK

		mov	temp, CNT
		add	temp, hundred_us
		waitcnt	temp, delay
		call 	#initialize

top_of_frame
:loop           mov     count, Hheight
                rdlong  highaddr, scanbase
                and     highaddr, ADDR_MASK
                mov     baaddr, highaddr
                and     baaddr, BAMASK
                mov     lowaddr, baaddr
                mov     VSCL, V25
:scanloop
                WAITVID refresh_comms, refresh_seq  ' synchronize with end of blanking
                call    #burst_read
                WAITVID refresh_comms, refresh_seq
                add     highaddr, #1
                and     highaddr, ADDR_MASK
                WAITVID refresh_comms, refresh_seq
                mov     baaddr, highaddr
                and     baaddr, BAMASK
                WAITVID refresh_comms, refresh_seq
                mov     lowaddr, baaddr
                djnz    count, #:scanloop

                WAITVID refresh_comms, refresh_seq
                mov     cycles_left, GAP_CYCLES
                mov     VSCL, Vwait
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
                jmp     #span_interp

BAMASK          long    $00006000
ADDR_MASK       long    $00007FFF

VIDMASK		long	$00FF0000
CKEMASK         long    $00400000
LCLKMASK	long	$02000000
pin_SCLK	long	15
pin_DCLK        long    24
SCLK_MASK	long	$00008000
DCLK_MASK       long    $01000000
vconfig_stop	long    %0_00_1_0_0_000_00000_000000010_011111111
vconfig_go	long    %0_01_1_0_0_000_00000_000000010_011111111



burst_read
                mov     VSCL, V800
                WAITVID read_comms1, read_seq1
                mov     OUTA, highaddr
                mov     OUTA, lowaddr
                mov     VSCL, V25
                WAITVID read_comms2, read_seq2
		mov     OUTA, baaddr ' precharge normal
burst_read_ret  ret


read_comms1     long    $DE564E5E  ' NOOP, ACTIVE, READ, NOP_DEN
read_seq1       long    %11_11_11_11_11_11_11_11_11_11_11_11_00_10_01_00

read_comms2     long    $DE4ADA5E  ' NOOP, BT_DEN, PRECHARGE, NOOP_DEN
read_seq2       long    %00_10_00_11_01_11_11



{
burst_write
		mov	OUTA, colour
                mov     temp, writelen
		add	temp, #4+1
		or	temp, V4096
		mov	VSCL, V5
		or	OUTA, LCLKMASK
		WAITVID write_comms0, write_seq0
                mov     OUTA, highaddr
                mov     OUTA, lowaddr
                WAITVID write_comms1, write_seq1
		mov	VSCL, temp	' writelen - 5
	        WAITVID write_comms1, write_seq1
		mov	VSCL, V25
		WAITVID write_comms2, write_seq2
                mov     OUTA, baaddr ' precharge normal
burst_write_ret ret
}

write_comms00   long    $5A504E5E  ' NOP, ACTIVE, WRITE_LE, BT
write_seq00     long    %11_10_00_01_00

write_comms0    long    $504E5C5E  ' NOOP, NOP_LE, ACTIVE, WRITE_LE
write_seq0      long    %01_11_00_10_00

write_comms1    long    $505A5C5E  ' NOP, NOP_LE, BT, WRITE_LE
write_seq1      long    %01_01_01_01_01_01_01_01_01_01_01_01_01_01_01_01  ' all NOP_LE
narrow_seq      long    %10_01_01_01_01        ' 4 NOP_LE's then BT

write_comms2    long    $4A5A5E  ' NOOP, BT, PRECHARGE
write_seq2      long    %00_10_01


char_setup      and     char, #$FF
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		ror	char, #1  wc		' Prop font address decoding, C flag has odd/even
                shl     char, #7
                add     char, fontbase

                mov     count, #32		' charline count

                mov     VSCL, V16		' 16 pixels wide
                or      OUTA, LCLKMASK
                WAITVID char_comms0, char_seq0   ' setup ACTIVE
                mov     OUTA, highaddr

                rdlong  patt, char		' first charline outside loop
	if_c    shr     patt, #1
                and     patt, H55555555
	  	or	patt, #2
		add	char, #4

:cloop
                mov     OUTA, lowaddr
                WAITVID char_comms1, patt

                rdlong  patt, char
		add	char, #4
	if_c    shr     patt, #1		' select odd or even character
                and     patt, H55555555	        ' mask out other
	  	or	patt, #2                ' insert WRITE bit
                add     highaddr, #1            ' ready for next row
		movs	VSCL, #7  		' for BT/PRECHARGE/ACTIVE

                WAITVID char_comms2, char_seq2  ' BT/PRECHARGE/NOP/ACTIVE
                mov     OUTA, baaddr            ' precharge
                movs    VSCL, #16
                mov     OUTA, highaddr          ' active

                djnz    count, #:cloop

                mov     VSCL, Vwait
                WAITVID char_comms2, char_seq2x
                mov     OUTA, baaddr     ' precharge
                sub     cycles_left, #16+16*(16+7)+WAITCYCLES   ' larger than 511
                sub     cycles_left, #16*(16+7)
char_setup_ret  ret


char            long    0
patt            long    0
fontbase        long    $8000
H55555555       long    $55555555
D2              long    2 << 9
V9              long    $1000 | 9
V16             long    $1000 | 16

char_comms0     long    $4E5E      ' NOOP, ACTIVE
char_seq0       long    %00_00_01_00

char_comms1     long    $50515C5D ' NOP_LE_DQM, NOP_LE, WR_LE_DQM, WR_LE

char_comms2     long    $4E4A5A5E   ' NOOP, BT, PRECHARGE, ACTIVE
char_seq2       long    %00_11_00_10_01
char_seq2x      long    %00_10_00_00_00


initialize
        	mov	VSCL, V7
 		WAITVID init_comms, init_seq
		mov	OUTA, precharge_all
                nop
                mov	OUTA, modeword
                mov     OUTA, #0
                mov     VSCL, V25 ' V4096
		WAITVID init_comms, #0
                mov     OUTA, #0
                mov     count, #410
:loop
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                WAITVID refresh_comms, refresh_seq
                djnz    count, #:loop
initialize_ret	ret


modeword	long	%00_0_00_010_0_111
precharge_all	long	$400		' A10 set


nop_comms       long    $5E5E5E5E
init_comms	long	$424A5E ' NOOP,PRECHARGE,LOADREG
init_seq	long	%00_10_00_01_00


refresh_comms   long    $464A5E ' NOOP, PRECHARGE, AUTO_REFR
refresh_seq     long    %00_10_00_00_10_00



span_interp	cmp	cycles_left, MIN_CYCLES  wz,wc
	if_b	jmp	#wait_frame          ' run out of time, we're done

		rdlong	w0, queuep
		cmp	w0, HIGH_WORD  wc,wz ' is slot empty, then we're done
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_ae	jmp	#span_interp

		add	queuep, #4
		rdlong	w1, queuep	     ' capture slot's data
		sub	queuep, #4
		wrlong	HIGH_WORD, queuep    ' mark slot empty

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		call	#handle_span

		add	queuep, #8
		cmp	queuep, queue_end  wz,wc
	if_ae	mov	queuep, queue_beg    ' wraparound

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
		jmp	#span_interp


wait_frame      cmps    cycles_left, #0  wz,wc
        if_be   jmp     #top_of_frame
                cmpsub  cycles_left, D2048  wc
        if_c    mov     VSCL, V2048
        if_nc   mov     VSCL, cycles_left
        if_nc   add     VSCL, WAITCYCLES
        if_nc   or      VSCL, V4096
        if_nc   mov     cycles_left, #0
                WAITVID refresh_comms, refresh_seq
                jmp     #wait_frame


handle_span	
		mov	command, w0
		shr	command, #16
		cmp	command, #COMM_COLOURS  wz
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_nz	jmp	#:not_col
		mov	colour, w0
		and	colour, #$FF  ' currently limited to 8 bit colour
		WAITVID refresh_comms, refresh_seq
		sub	cycles_left, #WAITCYCLES
		jmp	#handle_span_ret
:not_col
		mov	OUTA, colour    ' precharge address pins for latch
		mov	lowaddr, w0              ' a actual paint operation, handle x and y
		and	lowaddr, COLADDR_MASK
		mov	highaddr, w1
		shr	highaddr, #16
                and     highaddr, ADDR_MASK
                mov     baaddr, highaddr
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
                and     baaddr, BAMASK
                or      lowaddr, baaddr		  ' duplicate bank bits into column address

		cmp	command, #COMM_CHAR  wz
	if_nz	jmp	#:not_chr

		mov	char, w1
                and     char, #$FF		' font is 128 char one ATM
		call	#char_setup
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
		jmp	#handle_span_ret

:not_chr
		and	w1, LOW_WORD      ' actually this isn't quite correct since cannot handle a full 1024 width
		mov	writelen, w1            ' its a span, get the length
		cmp	writelen, #1  wz,wc
	if_be	jmp	#:onespan
		cmp	writelen, #5  wz,wc     ' test for very short spans
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_a	jmp	#:widespan
:narrowspan
		shl	w1, #1          ' calculate  2 * (5 - writelen)
		sub	w1, #10
		neg	w1, w1		' shift count to adapt
		mov	temp, narrow_seq ' narrow_seq is 4 NOP_LE's then BT
		shr	temp, w1	' shift down the sequence to cut out pixels
		mov	VSCL, V5
		or	OUTA, LCLKMASK
		WAITVID write_comms0, write_seq0  ' start the WRITE, does one pixel only
                mov     OUTA, highaddr
		nop
                mov     OUTA, lowaddr
                WAITVID write_comms1, temp   ' nops LE, poss BT, upto 3 more pixels
                mov     VSCL, Vwait
		WAITVID write_comms2, #%10_00 ' NOP, precharge
                mov     OUTA, baaddr ' precharge the right bank
                sub     cycles_left, #WAITCYCLES+5+5
                jmp     #handle_span_ret
:widespan
                mov     temp, writelen
                cmp     temp, #10  wz,wc
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
        if_a    jmp     #:fullspan

                sub     w1, #5		' handle from 6 to 10 pixels
		shl	w1, #1
		sub	w1, #10
		neg	w1, w1		' shift count to adapt, shift = 0 for 10 pixels, 2 for 9 pixels, ... 8 for 6 pixels
		mov	temp, narrow_seq 
		shr	temp, w1        ' handle between 0 and 4 pixels, for 6..10 incoming
		mov	VSCL, V5	' will do 6 anyway, 1 for write_seq0, 5 for write_seq1, then temp
		or	OUTA, LCLKMASK
		WAITVID write_comms0, write_seq0  ' start the WRITE
                mov     OUTA, highaddr
		nop
                mov     OUTA, lowaddr
                WAITVID write_comms1, write_seq1   ' nops LE
                WAITVID write_comms1, temp   ' nops LE, poss BT
                mov     OUTA, baaddr ' precharge normal
                mov     VSCL, Vwait
		WAITVID write_comms2, #%10 ' just precharge
                sub     cycles_left, #WAITCYCLES+5+5+5

                jmp     #handle_span_ret

:fullspan
		sub	temp, #6
                sub     cycles_left, temp
		mov	VSCL, V5
		or	OUTA, LCLKMASK
		WAITVID write_comms0, write_seq0
                mov     OUTA, highaddr
		nop
                mov     OUTA, lowaddr
	        WAITVID write_comms1, write_seq1
		or	temp, V4096
		mov	VSCL, temp	' writelen - 6
                WAITVID write_comms1, write_seq1
               	mov	VSCL, Vwait
		WAITVID write_comms2, write_seq2
                mov     OUTA, baaddr ' precharge normal
                sub     cycles_left, #WAITCYCLES+5+5
		jmp	#handle_span_ret

:onespan	mov	VSCL, V5
		or	OUTA, LCLKMASK
		WAITVID write_comms00, write_seq00  ' start the WRITE, does one pixel only
                mov     OUTA, highaddr
                mov     VSCL, Vwait
                mov     OUTA, lowaddr
		WAITVID write_comms2, #%10_00 ' just precharge
		mov	OUTA, baaddr
                sub     cycles_left, #WAITCYCLES+5

handle_span_ret ret


hundred_us	long	8000
delay		long	800

CLOCK_FREQ      long    $10000000  ' 5MHz * 16 / 4 = 20MHz PLL
PHASE           long    $00000000

' dataval		long	$000000FF ' 8 bit test pattern, may be 16bit later
writelen        long    250

DEN		long	%10000000 ' $80
LE		long	%00000010 ' $02
DQM		long	%00000001 ' $01

{
DEPIN		long	%10000000
CEPIN           long    %01000000
CSPIN           long    %00100000
RASPIN          long    %00010000
CASPIN          long    %00001000
WEPIN           long    %00000100
LEPIN		long	%00000010
DQMPIN		long	%00000001
}
DISABLED	long	%00111110 ' $3E
NOOP		long	%01011110 ' $5E
NOP_DEN		long	%11011110 ' $DE
NOP_LE		long	%01011100 ' $5C
NOP_LE_DQM	long	%01011101 ' $5D
NOP_DQM		long	%01011111 ' $5F
ACTIVE		long	%01001110 ' $4E
READ		long	%01010110 ' $56
WRITE		long	%01010010 ' $52
WRITE_LE	long	%01010000 ' $50
WRITE_LE_DQM	long	%01010001 ' $51
BT		long	%01011010 ' $5A
BT_DEN          long    %11011010 ' $DA
PRECHARGE	long	%01001010 ' $4A
AUTO_REFR	long	%01000110 ' $46
LOADREG		long	%01000010 ' $42

Hheight         long    HEIGHT
COLADDR_MASK    long    $3FF
'H8000           long    $8000
LOW_WORD        long    $0000FFFF
HIGH_WORD       long    $FFFF0000
D2048           long    $800
'H1FFF		long	$1FFF
'H1000           long    $1000

V4              long    $1000 | 4
V5              long    $1000 | 5
V6              long    $1000 | 6
V7              long    $1000 | 7
V11             long    $1000 | 11
V25             long    $1000 | 25
Vwait           long    $1000 | WAITCYCLES
V800            long    $1000 | 800
'V1024           long    $1000 | 1024
V2048           long    $1000 | 2048
V4096           long    $1000 | 4096

GAP_CYCLES      long    (WIDTH+GAP)*GAP
MIN_CYCLES      long    WIDTH+150
zero            long    0


parm            res     1
queuep          res     1
queue_beg       res     1
queue_end       res     1
temp            res     1
count           res     1
ref_count       res     1
w0              res     1
w1              res     1
colour          res     1
cycles_left     res     1
command         res     1

highaddr        res     1
lowaddr         res     1
baaddr          res     1
scanbase	res	1
lock		res	1

		FIT	$1F0

{{
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
}}
